#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# glumpy is an OpenGL framework for the fast visualization of numpy arrays.
# Copyright (C) 2009-2011  Nicolas P. Rougier. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY NICOLAS P. ROUGIER ''AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL NICOLAS P. ROUGIER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of Nicolas P. Rougier.
# -----------------------------------------------------------------------------
"""
A filter is a shader that transform the current displayed texture. Since
shaders cannot be easily serialized within the GPU, they have to be well
structured on the python side such that we can possibly merge them into a
single source code for both vertex and fragment. Consequently, there is a
default code for both vertex and fragment with specific entry points such that
filter knows where to insert their specific code (declarations, functions and
call (or code) to be inserted in the main function).

Spatial interpolation filter classes for OpenGL textures.

Each filter generates a one-dimensional lookup table (weights value from 0 to
ceil(radius)) that is uploaded to video memory (as a 1d texture) and is then
read by the shader when necessary. It avoids computing weight values for each
pixel. Furthemore, each 2D-convolution filter is separable and can be computed
using 2 1D-convolution with same 1d-kernel (= the lookup table values).

Available filters:

  - Nearest  (radius 0.5)
  - Linear (radius 1)
  - Hanning (radius 1)
  - Hamming (radius 1)
  - Hermite (radius 1)
  - Kaiser (radius 1)
  - Quadric (radius 1.5)
  - Cubic (radius 2)
  - CatRom (radius 2)
  - Mitchell (radius 2)
  - Spline16 (radius 2)
  - Spline36 (radius 4)
  - Gaussian (radius 2)
  - Bessel (radius 3.2383)
  - Sinc (radius 4)
  - Lanczos (radius 4)
  - Blackman (radius 4)


Note::

  Weights code has been translated from the antigrain geometry library
  available at http://www.antigrain.com/
"""


import math
import numpy as np
from inspect import cleandoc
from itertools import product


class SpatialFilter:
    def __init__(self, radius=1):
        self.radius = math.ceil(radius)

    def weight(self, x):
        """
        Return filter weight for a distance x.

        :Parameters:
            ``x`` : 0 < float < ceil(self.radius)
                Distance to be used to compute weight.
        """
        raise NotImplementedError

    def kernel(self, size=4 * 512):
        samples = int(size / self.radius)
        n = size  # r*samples
        kernel = np.zeros(n)
        X = np.linspace(0, self.radius, n)
        for i in range(n):
            kernel[i] = self.weight(X[i])
        N = np.zeros(samples)
        for i in range(self.radius):
            N += kernel[::+1][i * samples:(i + 1) * samples]
            N += kernel[::-1][i * samples:(i + 1) * samples]
        for i in range(self.radius):
            kernel[i * samples:(i + 1) * samples] /= N
        return kernel

    def call_code(self, index):
        code = cleandoc(f'''
            vec4 {self.__class__.__name__}2D(sampler2D texture, vec2 shape, vec2 uv) {{
                return filter2D_radius{self.radius}(texture, u_kernel, {index}, uv, 1 / shape);
            }}

            vec4 {self.__class__.__name__}3D(sampler3D texture, vec3 shape, vec3 uv) {{
                return filter3D_radius{self.radius}(texture, u_kernel, {index}, uv, 1 / shape);
            }}
        ''')

        return code


class Linear(SpatialFilter):
    """
    Linear filter (radius = 1).

    Weight function::

      w(x) = 1 - x

    """

    def weight(self, x):
        return 1 - x


class Hanning(SpatialFilter):
    """
    Hanning filter (radius = 1).

    Weight function::

      w(x) = 0.5 + 0.5 * cos(pi * x)

    """

    def weight(self, x):
        return 0.5 + 0.5 * math.cos(math.pi * x)


class Hamming(SpatialFilter):
    """
    Hamming filter (radius = 1).

    Weight function::

      w(x) = 0.54 + 0.46 * cos(pi * x)

    """

    def weight(self, x):
        return 0.54 + 0.46 * math.cos(math.pi * x)


class Hermite(SpatialFilter):
    """Hermite filter (radius = 1).

    Weight function::

      w(x) = (2*x-3)*x^2 + 1

    """

    def weight(self, x):
        return (2 * x - 3) * x**2 + 1


class Quadric(SpatialFilter):
    """
    Quadric filter (radius = 1.5).

    Weight function::

             |  0 ≤ x < 0.5: 0.75 - x*x
      w(x) = |  0.5 ≤ x < 1.5: 0.5 - (x-1.5)^2
             |  1.5 ≤ x      : 0

    """

    def __init__(self):
        super().__init__(radius=1.5)

    def weight(self, x):
        if x < 0.75:
            return 0.75 - x**2
        elif x < 1.5:
            t = x - 1.5
            return 0.5 * t**2
        return 0


class Cubic(SpatialFilter):
    """
    Cubic filter (radius = 2).

    Weight function::

      w(x) = 1/6((x+2)^3 - 4*(x+1)^3 + 6*x^3 -4*(x-1)^3)
    """

    def __init__(self):
        super().__init__(radius=2)

    def weight(self, x):
        return (1 / 6) * (
            (x + 2)**3
            - 4 * (x + 1)**3
            + 6 * x**3
            - 4 * (x - 1)**3
        )


class Kaiser(SpatialFilter):
    """
    Kaiser filter (radius = 1).


    Weight function::

      w(x) = bessel_i0(a sqrt(1-x^2)* 1/bessel_i0(b)

    """

    def __init__(self, b=6.33):
        self.a = b
        self.epsilon = 1e-12
        self.i0a = 1 / self.bessel_i0(b)
        super().__init__(radius=1)

    def bessel_i0(self, x):
        s = 1
        y = x**2 / 4
        t = y
        i = 2
        while t > self.epsilon:
            s += t
            t *= float(y) / i**2
            i += 1
        return s

    def weight(self, x):
        if x > 1:
            return 0
        return self.bessel_i0(self.a * math.sqrt(1 - x**2)) * self.i0a


class CatRom(SpatialFilter):
    """
    Catmull-Rom filter (radius = 2).

    Weight function::

             |  0 ≤ x < 1: 0.5*(2 + x^2*(-5+x*3))
      w(x) = |  1 ≤ x < 2: 0.5*(4 + x*(-8+x*(5-x)))
             |  2 ≤ x    : 0

    """

    def __init__(self):
        super().__init__(radius=2)

    def weight(self, x):
        if x < 1:
            return 0.5 * (2 + x**2 * (-5 + x * 3))
        elif x < 2:
            return 0.5 * (4 + x * (-8 + x * (5 - x)))
        else:
            return 0


class Mitchell(SpatialFilter):
    """
    Mitchell-Netravali filter (radius = 2).

    Weight function::

             |  0 ≤ x < 1: p0 + x^2*(p2 + x*p3)
      w(x) = |  1 ≤ x < 2: q0 + x*(q1 + x*(q2 + x*q3))
             |  2 ≤ x    : 0

    """

    def __init__(self, b=1/3, c=1/3):
        self.p0 = (6 - 2 * b) / 6
        self.p2 = (-18 + 12 * b + 6 * c) / 6
        self.p3 = (12 - 9 * b - 6 * c) / 6
        self.q0 = (8 * b + 24 * c) / 6
        self.q1 = (-12 * b - 48 * c) / 6
        self.q2 = (6 * b + 30 * c) / 6
        self.q3 = (-b - 6 * c) / 6
        super().__init__(radius=2)

    def weight(self, x):
        if x < 1:
            return self.p0 + x**2 * (self.p2 + x * self.p3)
        elif x < 2:
            return self.q0 + x * (self.q1 + x * (self.q2 + x * self.q3))
        else:
            return 0


class Spline16(SpatialFilter):
    """
    Spline16 filter (radius = 2).

    Weight function::

             |  0 ≤ x < 1: ((x-9/5)*x - 1/5)*x + 1
      w(x) = |
             |  1 ≤ x < 2: ((-1/3*(x-1) + 4/5)*(x-1) - 7/15 )*(x-1)

    """

    def __init__(self):
        super().__init__(radius=2)

    def weight(self, x):
        if x < 1:
            return ((x - 9/5) * x - 1/5) * x + 1
        else:
            return ((-1/3 * (x - 1) + 4/5) * (x - 1) - 7/15) * (x - 1)


class Spline36(SpatialFilter):
    """
    Spline36 filter (radius = 3).

    Weight function::

             |  0 ≤ x < 1: ((13/11*x - 453/209)*x -3/209)*x +1
      w(x) = |  1 ≤ x < 2: ((-6/11*(x-1) - 270/209)*(x-1) -156/209)*(x-1)
             |  2 ≤ x < 3: (( 1/11*(x-2) - 45/209)*(x-2) + 26/209)*(x-2)
    """

    def __init__(self):
        super().__init__(radius=3)

    def weight(self, x):
        if x < 1:
            return ((13/11 * x - 453/209) * x - 3/209) * x + 1
        elif x < 2:
            return ((-6/11 * (x - 1) + 270/209) * (x - 1) - 156 / 209) * (x - 1)
        else:
            return ((1/11 * (x - 2) - 45/209) * (x - 2) + 26/209) * (x - 2)


class Gaussian(SpatialFilter):
    """
    Gaussian filter (radius = 2).

    Weight function::

      w(x) = exp(-2x^2) * sqrt(2/pi)

    Note::

      This filter does not seem to be correct since:

        x = np.linspace(0, 1, 100 )
        f = weight
        z = f(x+1)+f(x)+f(1-x)+f(2-x)

        z should be 1 everywhere but it is not the case and it produces "grid
        effects".
    """

    def __init__(self):
        super().__init__(radius=2)

    def weight(self, x):
        return math.exp(-2 * x**2) * math.sqrt(2 / math.pi)


class Bessel(SpatialFilter):
    """Bessel filter (radius = 3.2383)."""

    def __init__(self):
        super().__init__(radius=3.2383)

    def besj(self, x, n):
        """Function BESJ calculates Bessel function of first kind of order n.

        Parameters
        ----------
        x: int
            value at which the Bessel function is required
        n : int
            an integer (>=0), the order

        Notes
        -----
        C++ Mathematical Library
        Converted from equivalent FORTRAN library
        Converted by Gareth Walker for use by course 392 computational project
        All functions tested and yield the same results as the corresponding
        FORTRAN versions.

        If you have any problems using these functions please report them to
        M.Muldoon@UMIST.ac.uk

        Documentation available on the web
        http://www.ma.umist.ac.uk/mrm/Teaching/392/libs/392.html
        Version 1.0   8/98
        29 October, 1999

        Adapted for use in AGG library by
                    Andy Wilk (castor.vulgaris@gmail.com)
        Adapted for use in vispy library by
                    Nicolas P. Rougier (Nicolas.Rougier@inria.fr)

        """
        if n < 0:
            return 0
        x = float(x)  # force float type

        d = 1e-6
        b = 0
        if math.fabs(x) <= d:
            if n != 0:
                return 0
            return 1

        b1 = 0  # b1 is the value from the previous iteration
        # Set up a starting order for recurrence
        m1 = int(math.fabs(x)) + 6
        if math.fabs(x) > 5:
            m1 = int(math.fabs(1.4 * x + 60 / x))

        m2 = int(n + 2 + math.fabs(x) / 4)
        if m1 > m2:
            m2 = m1

        # Apply recurrence down from current max order
        while True:
            c3 = 0
            c2 = 1e-30
            c4 = 0
            m8 = 1
            if m2 // 2 * 2 == m2:
                m8 = -1

            imax = m2 - 2
            for i in range(1, imax + 1):
                c6 = 2 * (m2 - i) * c2 / x - c3
                c3 = c2
                c2 = c6
                if m2 - i - 1 == n:
                    b = c6
                m8 = -1 * m8
                if m8 > 0:
                    c4 = c4 + 2 * c6

            c6 = 2 * c2 / x - c3
            if n == 0:
                b = c6
            c4 += c6
            b /= c4
            if math.fabs(b - b1) < d:
                return b
            b1 = b
            m2 += 3

    def weight(self, x):
        if x == 0:
            return math.pi / 4
        else:
            return self.besj(math.pi * x, 1) / (2 * x)


class Sinc(SpatialFilter):
    """Sinc filter (radius = 4)."""

    def __init__(self):
        super().__init__(radius=4)

    def weight(self, x):
        if x == 0:
            return 1
        x *= math.pi
        return (math.sin(x) / x)


class Lanczos(SpatialFilter):
    """Lanczos filter (radius = 4)."""

    def __init__(self):
        super().__init__(radius=4)

    def weight(self, x):
        if x == 0:
            return 1
        elif x > self.radius:
            return 0
        x *= math.pi
        xr = x / self.radius
        return (math.sin(x) / x) * (math.sin(xr) / xr)


class Blackman(SpatialFilter):
    """Blackman filter (radius = 4)."""

    def __init__(self):
        super().__init__(radius=4)

    def weight(self, x):
        if x == 0:
            return 1
        elif x > self.radius:
            return 0
        x *= math.pi
        xr = x / self.radius
        return (math.sin(x) / x) * (0.42 + 0.5 * math.cos(xr) + 0.08 * math.cos(2 * xr))


def generate_filter_code(radius):
    n = int(math.ceil(radius))

    nl = '\n'  # cannot use backslash in fstring
    code = cleandoc(f'''
    vec4 filter1D_radius{n}(sampler2D kernel, float index, float x{''.join(f', vec4 c{i}' for i in range(n * 2))}) {{
        float w, w_sum = 0;
        vec4 r = vec4(0);
        {''.join(f"""
        w = unpack_interpolate(kernel, vec2({1 - (i + 1) / n} + (x / {n}), index));
        w = w * kernel_scale + kernel_bias;
        r += c{i} * w;
        w = unpack_interpolate(kernel, vec2({(i + 1) / n} - (x / {n}), index));
        w = w * kernel_scale + kernel_bias;
        r += c{i + n} * w;"""
        for i in range(n))}
        return r;
    }}

    vec4 filter2D_radius{n}(sampler2D texture, sampler2D kernel, float index, vec2 uv, vec2 pixel) {{
        vec2 texel = uv / pixel - vec2(0.5);
        vec2 f = fract(texel);
        texel = (texel - fract(texel) + vec2(0.001)) * pixel;
        {''.join(f"""
        vec4 t{i} = filter1D_radius{n}(kernel, index, f.x{f''.join(
            f',{nl}            texture2D(texture, texel + vec2({-n + 1 + j}, {-n + 1 + i}) * pixel)'
            for j in range(n * 2))});"""
        for i in range(n * 2))}
        return filter1D_radius{n}(kernel, index, f.y{''.join(f', t{i}' for i in range(2*n))});
    }}

    vec4 filter3D_radius{n}(sampler3D texture, sampler2D kernel, float index, vec3 uv, vec3 pixel) {{
        vec3 texel = uv / pixel - vec3(0.5);
        vec3 f = fract(texel);
        texel = (texel - fract(texel) + vec3(0.001)) * pixel;
        {''.join(f"""
        vec4 t{i}{j} = filter1D_radius{n}(kernel, index, f.x{f''.join(
            f',{nl}            texture3D(texture, texel + vec3({-n + 1 + k}, {-n + 1 + j}, {-n + 1 + i}) * pixel)'
            for k in range(n * 2))});"""
        for i, j in product(range(n * 2), range(n * 2)))}
        {f''.join(f"""
        vec4 t{i} = filter1D_radius{n}(kernel, index, f.y{"".join(
            f", t{i}{j}" for j in range(n * 2))});"""
        for i in range(n * 2))}
        return filter1D_radius{n}(kernel, index, f.z{''.join(f', t{i}' for i in range(2*n))});
    }}
    ''')

    return code


def main():
    # Generate kernels texture (16 x 1024)
    filters = [Linear(), Hanning(), Hamming(), Hermite(), Kaiser(), Quadric(),
               Cubic(), CatRom(), Mitchell(), Spline16(), Spline36(), Gaussian(),
               Bessel(), Sinc(), Lanczos(), Blackman()]

    n = 1024
    K = np.zeros((len(filters), n))
    for i, f in enumerate(filters):
        K[i] = f.kernel(n)

    bias = K.min()
    scale = K.max() - K.min()
    K = (K - bias) / scale
    np.save("spatial-filters.npy", K.astype(np.float32))

    code = cleandoc(f'''
        // ------------------------------------
        // Automatically generated, do not edit
        // ------------------------------------
        const float kernel_bias  = {bias};
        const float kernel_scale = {scale};
        const float kernel_size = {n};
        const vec4 bits = vec4(1, {1 / 256}, {1 / (256 * 256)}, {1 / (256 * 256 * 256)});
        uniform sampler2D u_kernel;
    ''')

    # add basic unpack functions
    code += '\n\n' + cleandoc('''
        float unpack_unit(vec4 rgba) {
            // return rgba.r;  // uncomment this for r32f debugging
            return dot(rgba, bits);
        }

        float unpack_ieee(vec4 rgba) {
            // return rgba.r;  // uncomment this for r32f debugging
            rgba.rgba = rgba.abgr * 255;
            float sign = 1 - step(128 , rgba[0]) * 2;
            float exponent = 2 * mod(rgba[0] , 128) + step(128 , rgba[1]) - 127;
            float mantissa = mod(rgba[1] , 128) * 65536 + rgba[2] * 256 + rgba[3] + float(0x800000);
            return sign * exp2(exponent) * (mantissa * exp2(-23.));
        }


        float unpack_interpolate(sampler2D kernel, vec2 uv) {
            // return texture2D(kernel, uv).r;  //uncomment this for r32f debug without interpolation
            float kpixel = 1. / kernel_size;
            float u = uv.x / kpixel;
            float v = uv.y;
            float uf = fract(u);
            u = (u - uf) * kpixel;
            float d0 = unpack_unit(texture2D(kernel, vec2(u, v)));
            float d1 = unpack_unit(texture2D(kernel, vec2(u + 1. * kpixel, v)));
            return mix(d0, d1, uf);
        }
    ''')

    # add 1d, 2d and 3d filter code
    for radius in range(4):
        code += '\n\n' + generate_filter_code(radius + 1)

    # add call functions for 2D and 3D filters
    # special case for nearest
    code += '\n\n' + cleandoc('''
        vec4 Nearest2D(sampler2D texture, vec2 shape, vec2 uv) {
            return texture2D(texture, uv);
        }

        vec4 Nearest3D(sampler3D texture, vec3 shape, vec3 uv) {
            return texture3D(texture, uv);
        }
    ''')

    for i, f in enumerate(filters):
        code += '\n\n' + f.call_code((i + 0.5) / 16)

    print(code)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Utils for offscreen rendering.
"""

from .. import Application, Canvas
from ... import gloo


class OffscreenContext:
    """ A helper class to provide an OpenGL context. This context is global
    to the application.
    """

    _global_instance = None
    _canvas = None

    @classmethod
    def get_global_instance(cls):
        """ Get a global context. Note that any assumptions about OpenGL state
        being local will not hold.
        """
        if cls._global_instance is None:
            cls._global_instance = cls()
        return cls._global_instance

    def __init__(self):
        if self._canvas is not None:
            return  # already initialized

        self._is_closed = False

        # Glfw is probably the most lightweight approach, so let's try that.
        # But there are two incompatible packages providing glfw :/
        self.glfw = None
        try:
            import glfw
        except ImportError:
            pass
        else:
            need_from_glfw = ["create_window", "make_context_current"]
            if all(hasattr(glfw, attr) for attr in need_from_glfw):
                self.glfw = glfw

        if self.glfw:
            self.glfw.init()
            self.glfw.window_hint(self.glfw.VISIBLE, 0)
            self._canvas = self.glfw.create_window(1, 1, "dummy window", None, None)
        else:
            try:
                _app = Application('default')
            except Exception:
                raise RuntimeError(
                    "Cannot find a backend to create an OpenGL context. "
                    "Install e.g. PyQt5, PySide2, or `pip install glfw`."
                )
            self._canvas = Canvas(app=_app)
            self._canvas.show(False)

    def make_current(self):
        """ Make this the currently active context.
        """
        # If an application only used off-screen canvases this would technically
        # have to be called just once. But note that an application/session
        # could run both real canvases and off-screen ones.
        if self.glfw:
            self.glfw.make_context_current(self._canvas)
        else:
            self._canvas.set_current()

    def close(self):
        """ Close the context. """
        # Cannot close the global instance
        if self is OffscreenContext._global_instance:
            return
        elif not self._is_closed:
            self._is_closed = True
            if self.glfw:
                self.glfw.destroy_window(self._canvas)
            else:
                self._canvas.close()

    def __del__(self):
        self.close()


class FrameBufferHelper:
    """ Provides a canvas to render to, using an FBO.
    """

    def __init__(self):
        self._fbo = None
        self._physical_size = 1, 1
        self._fbo_size = -1, -1

    def _ensure_fbo(self):
        if self._fbo_size != self._physical_size:
            self._fbo_size = self._physical_size
            w, h = self._fbo_size
            if self._fbo is None:
                color_buffer = gloo.Texture2D((h, w, 4))
                depth_buffer = gloo.RenderBuffer((h, w))
                self._fbo = gloo.FrameBuffer(color_buffer, depth_buffer)
            else:
                self._fbo.resize((h, w))

    def set_physical_size(self, w, h):
        """ Set the physical size of the canvas.
        """
        self._physical_size = w, h

    def get_frame(self):
        """ Call this within the with-context to obtain the frame buffer contents.
        """
        return self._fbo.read()

    def __enter__(self):
        self._ensure_fbo()
        return self._fbo.__enter__()

    def __exit__(self, *args):
        return self._fbo.__exit__(*args)



from vispy.app.backends._offscreen_util import OffscreenContext, FrameBufferHelper
from vispy.testing import run_tests_if_main, requires_application
from vispy import gloo
import numpy as np


@requires_application()
def test_offscreen_context():
    c1 = OffscreenContext()
    c2 = OffscreenContext.get_global_instance()
    c3 = OffscreenContext.get_global_instance()
    c4 = OffscreenContext()

    assert c1 is not c2
    assert c1 is not c4
    assert c2 is c3

    c1.make_current()
    c1.close()


class FakeCanvas(object):

    def __init__(self):
        self.context = gloo.GLContext()
        gloo.context.set_current_canvas(self)

    def flush(self):
        self.context.flush_commands()


@requires_application()
def test_frame_buffer_helper():
    canvas = FakeCanvas()
    gl_context = OffscreenContext()
    fbh = FrameBufferHelper()

    gl_context.make_current()
    fbh.set_physical_size(43, 67)
    with fbh:
        gloo.set_clear_color((0, 0.5, 1))
        gloo.clear()
        canvas.flush()
        array = fbh.get_frame()

    assert array.shape == (67, 43, 4)
    assert np.all(array[:, :, 0] == 0)
    assert np.all(array[:, :, 1] == 128)
    assert np.all(array[:, :, 2] == 255)


run_tests_if_main()



# -*- coding: utf-8 -*-
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
"""Test the sphinx-gallery custom scraper."""

import os

from vispy.testing import TestingCanvas, requires_application

import pytest

try:
    from sphinx_gallery.gen_gallery import DEFAULT_GALLERY_CONF
except ImportError:
    pytest.skip("Skipping sphinx-gallery tests", allow_module_level=True)

from ..gallery_scraper import VisPyGalleryScraper


def _create_fake_block_vars(canvas):
    block_vars = {
        "example_globals": {
            "canvas": canvas,
        },
        "src_file": "example.py",
        "image_path_iterator": (f"{x}.png" for x in range(10)),
    }
    return block_vars


def _create_fake_gallery_conf(src_dir):
    gallery_conf = {}
    gallery_conf.update(DEFAULT_GALLERY_CONF)
    gallery_conf.update({
        "compress_images": "images",
        "compress_images_args": [],
        "src_dir": src_dir,
        "gallery_dirs": src_dir,
    })
    return gallery_conf


@requires_application()
@pytest.mark.parametrize("include_gallery_comment", [False, True])
def test_single_frame(include_gallery_comment, tmpdir):
    canvas = TestingCanvas()
    block_vars = _create_fake_block_vars(canvas)
    gallery_conf = _create_fake_gallery_conf(str(tmpdir))
    script = "\n# vispy: gallery 30\n" if include_gallery_comment else ""
    with tmpdir.as_cwd():
        with open("example.py", "w") as example_file:
            example_file.write(script)
        scraper = VisPyGalleryScraper()
        rst = scraper(None, block_vars, gallery_conf)
        if include_gallery_comment:
            assert "0.png" in rst
            assert os.path.isfile("0.png")
        else:
            assert "0.png" not in rst
            assert not os.path.isfile("0.png")
        assert not os.path.isfile("1.png")  # only one file created


@requires_application()
def test_single_animation(tmpdir):
    canvas = TestingCanvas()
    block_vars = _create_fake_block_vars(canvas)
    gallery_conf = _create_fake_gallery_conf(str(tmpdir))
    with tmpdir.as_cwd():
        with open("example.py", "w") as example_file:
            example_file.write("""# vispy: gallery 10:50:5
            """)
        scraper = VisPyGalleryScraper()
        rst = scraper(None, block_vars, gallery_conf)
        assert "0.gif" in rst
        assert os.path.isfile("0.gif")
        assert not os.path.isfile("0.png")  # only gif file created
        assert not os.path.isfile("1.png")  # only one file created


@requires_application()
@pytest.mark.parametrize(
    "exported_files",
    [
        ("example.png",),
        ("example.gif",),
        ("example1.png", "example2.png"),
    ])
def test_single_export(exported_files, tmpdir):
    canvas = TestingCanvas()
    block_vars = _create_fake_block_vars(canvas)
    gallery_conf = _create_fake_gallery_conf(str(tmpdir))
    with tmpdir.as_cwd():
        # create the files that the example should have created
        for fn in exported_files:
            open(fn, "w").close()

        with open("example.py", "w") as example_file:
            example_file.write("""# vispy: gallery-exports {}
            """.format(" ".join(exported_files)))

        scraper = VisPyGalleryScraper()
        rst = scraper(None, block_vars, gallery_conf)
        for idx, fn in enumerate(exported_files):
            # the original file should have been moved
            assert not os.path.isfile(fn)
            # the new name should from the sphinx-gallery iterator
            new_name = str(idx) + os.path.splitext(fn)[1]
            assert os.path.isfile(new_name)
            assert new_name in rst



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
import numpy as np
import pytest
from vispy.scene.cameras import TurntableCamera
from vispy.testing import run_tests_if_main


@pytest.mark.parametrize(
    "elevation, azimuth, roll, expected",
    [
        [0, 0, 0, np.eye(4)],
        [90, 0, 0, [[1, 0, 0, 0], [0, 0, -1, 0], [0, 1, 0, 0], [0, 0, 0, 1]]],
        [0, 90, 0, [[0, 1, 0, 0], [-1, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]],
        [0, 0, 90, [[0, 0, -1, 0], [0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 0, 1]]],
    ],
)
def test_turntable_camera_transform(elevation, azimuth, roll, expected):
    camera = TurntableCamera(elevation=elevation, azimuth=azimuth, roll=roll)
    matrix = camera._get_rotation_tr()
    np.testing.assert_allclose(matrix, expected, atol=1e-5)


run_tests_if_main()



# -*- coding: utf-8 -*- 
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

from vispy.visuals.mesh import MeshVisual
from vispy.visuals.filters import WireframeFilter


def test_empty_mesh_wireframe():
    """Test that an empty mesh does not cause issues with wireframe filter"""
    mesh = MeshVisual()
    wf = WireframeFilter()
    mesh.attach(wf)

    # trigger update of filter
    wf.enabled = True



# -*- coding: utf-8 -*-
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

from vispy import gloo, scene
from vispy.testing import requires_application, TestingCanvas
from vispy.visuals.transforms import STTransform

import numpy as np
import pytest


@requires_application()
@pytest.mark.parametrize(
    'blend_func',
    [
        ('src_alpha', 'one_minus_src_alpha', 'one', 'one_minus_src_alpha'),
        ('src_alpha', 'one_minus_src_alpha'),
        None,
    ])
def test_canvas_render(blend_func):
    """Test rendering a canvas to an array.

    Different blending functions are used to test what various Visuals may
    produce without actually using different types of Visuals.

    """
    with TestingCanvas(size=(125, 125), show=True, title='run') as c:
        view = c.central_widget.add_view()

        im1 = np.zeros((100, 100, 4)).astype(np.float32)
        im1[:, :, 0] = 1
        im1[:, :, 3] = 1

        im2 = np.zeros((50, 50, 4)).astype(np.float32)
        im2[:, :, 1] = 1
        im2[:, :, 3] = 0.4

        # Create the image
        image1 = scene.visuals.Image(im1, parent=view.scene)
        image1.transform = STTransform(translate=(20, 20, 0))
        image2 = scene.visuals.Image(im2, parent=view.scene)
        image2.transform = STTransform(translate=(0, 0, -1))
        if blend_func:
            image1.set_gl_state(preset='translucent', blend_func=blend_func)
            image2.set_gl_state(preset='translucent', blend_func=blend_func)

        rgba_result = c.render()
        rgb_result = c.render(alpha=False)

        # the results should be the same except for alpha
        np.testing.assert_allclose(rgba_result[..., :3], rgb_result)
        # the image should have something drawn in it
        assert not np.allclose(rgba_result[..., :3], 0)
        # the alpha should not be completely transparent
        assert not np.allclose(rgba_result[..., 3], 0)
        if blend_func is None or 'one' in blend_func:
            # no transparency
            np.testing.assert_allclose(rgba_result[..., 3], 255)
        else:
            # the alpha should have some transparency
            assert (rgba_result[..., 3] != 255).any()


@requires_application()
def test_picking_basic():
    """Test basic picking behavior.

    Based on https://github.com/vispy/vispy/issues/2107.

    """
    with TestingCanvas(size=(125, 125), show=True, title='run') as c:
        view = c.central_widget.add_view()
        view.margin = 5  # add empty space where there are no visuals
        view.camera = 'panzoom'

        x = np.linspace(0, 400, 100)
        y = np.linspace(0, 200, 100)
        line = scene.Line(np.array((x, y)).T.astype(np.float32))
        line.interactive = True
        view.add(line)
        view.camera.set_range()

        c.render()  # initial basic draw
        for _ in range(2):  # run picking twice to make sure it is repeatable
            # get Visuals on a Canvas point that Line is drawn on
            picked_visuals = c.visuals_at((100, 25))
            assert len(picked_visuals) == 2
            assert any(isinstance(vis, scene.ViewBox) for vis in picked_visuals)
            assert any(isinstance(vis, scene.Line) for vis in picked_visuals)


@requires_application()
@pytest.mark.parametrize(
    'preset',
    [
        'opaque', 'additive', 'translucent',
    ])
def test_blend_presets(preset):
    """Test blending presets render a canvas to an array.

    Different blending presets are used to test that they properly set
    blend equations.

    """
    with TestingCanvas(size=(125, 125), show=True, title='run') as c:
        view = c.central_widget.add_view()
        im1 = np.zeros((100, 100, 4)).astype(np.float32)
        im1[:, :, 1] = 1
        im1[:, :, 3] = .4
        # Create the image
        image1 = scene.visuals.Image(im1, parent=view.scene)
        image1.transform = STTransform(translate=(20, 20, -1))

        gloo.set_state(blend_equation='min')
        image1.set_gl_state(preset)

        rgba_result = c.render()
        assert not np.allclose(rgba_result[..., :3], 0)



# -*- coding: utf-8 -*-
import numpy as np
import pytest
from vispy.scene.visuals import Markers
from vispy.testing import (requires_application, TestingCanvas,
                           run_tests_if_main)
from vispy.testing.image_tester import assert_image_approved


@requires_application()
def test_markers():
    """Test basic marker / point-sprite support"""
    # this is probably too basic, but it at least ensures that point sprites
    # work for people
    np.random.seed(57983)
    data = np.random.normal(size=(30, 2), loc=50, scale=10)

    with TestingCanvas() as c:
        marker = Markers(parent=c.scene)
        marker.set_data(data)
        assert_image_approved(c.render(), "visuals/markers.png")

    # Test good correlation at high-dpi
    with TestingCanvas(px_scale=2) as c:
        marker = Markers(parent=c.scene)
        marker.set_data(data)
        assert_image_approved(c.render(), "visuals/markers.png")


def test_markers_edge_width():
    data = np.random.rand(10, 3)
    edge_width = np.random.rand(10)
    marker = Markers()

    with pytest.raises(ValueError):
        marker.set_data(pos=data, edge_width_rel=1, edge_width=1)

    marker.set_data(pos=data, edge_width=2)
    marker.set_data(pos=data, edge_width=edge_width)
    with pytest.raises(ValueError):
        marker.set_data(pos=data, edge_width=-1)

    marker.set_data(pos=data, edge_width_rel=edge_width, edge_width=None)
    marker.set_data(pos=data, edge_width_rel=edge_width + 1, edge_width=None)
    with pytest.raises(ValueError):
        marker.set_data(pos=data, edge_width_rel=edge_width - 1, edge_width=None)


def test_empty_markers_symbol():
    markers = Markers()
    markers.symbol = 'o'


run_tests_if_main()



# -*- coding: utf-8 -*-
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
"""Scraper for sphinx-gallery.

This is used to collect screenshots from the examples when executed via
sphinx-gallery. This can be included in any project wanting to take
advantage of this by adding the following to your sphinx ``conf.py``:

.. code-block:: python

    sphinx_gallery_conf = {
        ...
        'image_scrapers': ('vispy',)
    }

The scraper is provided to sphinx-gallery via the
``vispy._get_sg_image_scraper()`` function.

"""

from __future__ import annotations

import os
import time
import shutil
from vispy.io import imsave
from vispy.gloo.util import _screenshot
from vispy.scene import SceneCanvas

from sphinx_gallery.scrapers import optipng, figure_rst


class VisPyGalleryScraper:
    """Custom sphinx-gallery scraper to save the current Canvas to an image."""

    def __repr__(self):
        return self.__class__.__name__

    def __call__(self, block, block_vars, gallery_conf):
        """Scrape VisPy Canvases and applications.

        Parameters
        ----------
        block : tuple
            A tuple containing the (label, content, line_number) of the block.
        block_vars : dict
            Dict of block variables.
        gallery_conf : dict
            Contains the configuration of Sphinx-Gallery

        Returns
        -------
        rst : str
            The ReSTructuredText that will be rendered to HTML containing
            the images. This is often produced by
            :func:`sphinx_gallery.scrapers.figure_rst`.

        """
        example_fn = block_vars["src_file"]
        frame_num_list = self._get_frame_list_from_source(example_fn)
        image_path_iterator = block_vars['image_path_iterator']
        canvas_or_widget = get_canvaslike_from_globals(block_vars["example_globals"])
        if not frame_num_list:
            image_paths = []
        elif isinstance(frame_num_list[0], str):
            # example produces an image/animation as output
            image_paths = []
            for frame_image, image_path in zip(frame_num_list, image_path_iterator):
                image_path = os.path.splitext(image_path)[0] + os.path.splitext(frame_image)[1]
                shutil.move(frame_image, image_path)
                image_paths.append(image_path)
        else:
            image_paths = self._save_example_to_files(
                canvas_or_widget, frame_num_list, gallery_conf, image_path_iterator)

        fig_titles = ""  # alt text
        # FUTURE: Handle non-images (ex. MP4s) with raw HTML
        return figure_rst(image_paths, gallery_conf['src_dir'], fig_titles)

    def _save_example_to_files(self, canvas_or_widget, frame_num_list, gallery_conf, image_path_iterator):
        image_path = next(image_path_iterator)
        frame_grabber = FrameGrabber(canvas_or_widget, frame_num_list)
        frame_grabber.collect_frames()
        if len(frame_num_list) > 1:
            # let's make an animation
            # FUTURE: mp4 with imageio?
            image_path = os.path.splitext(image_path)[0] + ".gif"
            frame_grabber.save_animation(image_path)
        else:
            frame_grabber.save_frame(image_path)
        if 'images' in gallery_conf['compress_images']:
            optipng(image_path, gallery_conf['compress_images_args'])
        return [image_path]

    def _get_frame_list_from_source(self, filename):
        lines = open(filename, 'rb').read().decode('utf-8').splitlines()
        for line in lines[:10]:
            if not line.startswith("# vispy:"):
                continue
            if "gallery-exports" in line:
                _frames = line.split('gallery-exports')[1].split(',')[0].strip()
                frames = self._frame_exports_to_list(_frames)
                break
            if "gallery " in line:
                # Get what frames to grab
                _frames = line.split('gallery')[1].split(',')[0].strip()
                frames = self._frame_specifier_to_list(_frames)
                break
        else:
            # no frame number hint - don't grab any frames
            frames = []
        return frames

    def _frame_specifier_to_list(self, frame_specifier):
        _frames = frame_specifier or '0'
        frames = [int(i) for i in _frames.split(':')]
        if not frames:
            frames = [5]
        if len(frames) > 1:
            frames = list(range(*frames))
        return frames

    def _frame_exports_to_list(self, frame_specifier):
        frames = frame_specifier.split(" ")
        frame_paths = []
        for frame_fn in frames:
            # existing image file created by the example
            if not os.path.isfile(frame_fn):
                raise FileNotFoundError(
                    "Example gallery frame specifier must be a frame number, "
                    "frame range, or relative filename produced by the example.")
            frame_paths.append(frame_fn)
        return frame_paths


def get_canvaslike_from_globals(globals_dict):
    qt_widget = _get_qt_top_parent(globals_dict)
    if qt_widget is not None:
        return qt_widget

    # Get canvas
    if "canvas" in globals_dict:
        return globals_dict["canvas"]
    if "Canvas" in globals_dict:
        return globals_dict["Canvas"]()
    if "fig" in globals_dict:
        return globals_dict["fig"]
    return None


def _get_qt_top_parent(globals_dict):
    if "QWidget" not in globals_dict and "QMainWindow" not in globals_dict and "QtWidgets" not in globals_dict:
        return None

    qtwidgets = globals_dict.get("QtWidgets")
    qmainwindow = globals_dict.get("QMainWindow", getattr(qtwidgets, "QMainWindow", None))
    qwidget = globals_dict.get("QWidget", getattr(qtwidgets, "QWidget", qmainwindow))
    all_qt_widgets = [widget for widget in globals_dict.values()
                      if isinstance(widget, qwidget) and widget is not None]
    all_qt_mains = [widget for widget in all_qt_widgets if isinstance(widget, qmainwindow)]
    if all_qt_mains:
        return all_qt_mains[0]
    if all_qt_widgets:
        return all_qt_widgets[0]
    return None


class FrameGrabber:
    """Helper to grab a series of screenshots from the current Canvas-like object."""

    def __init__(self, canvas_obj, frame_grab_list: list[int]):
        self._canvas = canvas_obj
        self._done = False
        self._current_frame = -1
        self._collected_images = []
        self._frames_to_grab = frame_grab_list[:]  # copy so original list is preserved

    def on_draw(self, _):
        if self._done:
            return  # Grab only once
        self._current_frame += 1
        if self._current_frame in self._frames_to_grab:
            self._frames_to_grab.remove(self._current_frame)
            if isinstance(self._canvas, SceneCanvas):
                im = self._canvas.render(alpha=True)
            else:
                im = _screenshot()
            self._collected_images.append(im)
        if not self._frames_to_grab or self._current_frame > self._frames_to_grab[0]:
            self._done = True

    def collect_frames(self):
        """Show current Canvas and render and collect all frames requested."""
        if self._is_qt_widget():
            self._grab_qt_screenshot()
        else:
            self._grab_vispy_screenshots()

    def _is_qt_widget(self):
        try:
            from PyQt5.QtWidgets import QWidget
        except ImportError:
            return False
        return isinstance(self._canvas, QWidget)

    def _grab_qt_screenshot(self):
        from PyQt5.QtWidgets import QApplication
        from PyQt5.QtCore import QTimer
        self._canvas.show()
        # Qt is going to grab from the screen so we need the window on top
        self._canvas.raise_()
        # We need to give the GUI event loop and OS time to draw everything
        time.sleep(1.5)
        QApplication.processEvents()
        QTimer.singleShot(1000, self._grab_widget_screenshot)
        time.sleep(1.5)
        QApplication.processEvents()

    def _grab_widget_screenshot(self):
        from PyQt5.QtWidgets import QApplication
        screen = QApplication.screenAt(self._canvas.pos())
        screenshot = screen.grabWindow(int(self._canvas.windowHandle().winId()))
        arr = self._qpixmap_to_ndarray(screenshot)
        self._collected_images.append(arr)

    @staticmethod
    def _qpixmap_to_ndarray(pixmap):
        from PyQt5 import QtGui
        import numpy as np
        im = pixmap.toImage().convertToFormat(QtGui.QImage.Format.Format_RGB32)
        size = pixmap.size()
        width = size.width()
        height = size.height()
        im_bits = im.constBits()
        im_bits.setsize(height * width * 4)
        # Convert 0xffRRGGBB buffer -> (B, G, R, 0xff) -> (R, G, B)
        return np.array(im_bits).reshape((height, width, 4))[:, :, 2::-1]

    def _grab_vispy_screenshots(self):
        os.environ['VISPY_IGNORE_OLD_VERSION'] = 'true'
        self._canvas.events.draw.connect(self.on_draw, position='last')
        with self._canvas as c:
            self._collect_frames(c)

    def _collect_frames(self, canvas, limit=10000):
        n = 0
        while not self._done and n < limit:
            canvas.update()
            canvas.app.process_events()
            n += 1
        if n >= limit or len(self._frames_to_grab) > 0:
            raise RuntimeError("Could not collect any images")

    def save_frame(self, filename, frame_index=0):
        imsave(filename, self._collected_images[frame_index])

    def save_animation(self, filename):
        import imageio  # multiple gif not properly supported yet
        imageio.mimsave(filename, self._collected_images)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
"""Marker Visual and shader definitions."""

import numpy as np

from ..color import ColorArray
from ..gloo import VertexBuffer
from .shaders import Function, Variable
from .visual import Visual


_VERTEX_SHADER = """
uniform float u_antialias;
uniform float u_px_scale;
uniform bool u_scaling;
uniform bool u_spherical;

attribute vec3 a_position;
attribute vec4 a_fg_color;
attribute vec4 a_bg_color;
attribute float a_edgewidth;
attribute float a_size;
attribute float a_symbol;

varying vec4 v_fg_color;
varying vec4 v_bg_color;
varying float v_edgewidth;
varying float v_depth_middle;
varying float v_alias_ratio;
varying float v_symbol;

float big_float = 1e10; // prevents numerical imprecision

void main (void) {
    v_fg_color  = a_fg_color;
    v_bg_color  = a_bg_color;
    // fluctuations can mess "fake integers" up, so we do +0.5 and floor to make sure it's right
    v_symbol = a_symbol + 0.5;

    vec4 pos = vec4(a_position, 1);
    vec4 fb_pos = $visual_to_framebuffer(pos);
    gl_Position = $framebuffer_to_render(fb_pos);

    // NOTE: gl_stuff uses framebuffer coords!

    if (u_scaling == true) {
        // calculate point size from visual to framebuffer coords to determine size
        // move horizontally in framebuffer space
        // then go to scene coordinates (not visual, so scaling is accounted for)
        vec4 x = $framebuffer_to_scene(fb_pos + vec4(big_float, 0, 0, 0));
        // subtract position, so we get the scene-coordinate vector describing
        // an "horizontal direction parallel to the screen"
        vec4 scene_pos = $framebuffer_to_scene(fb_pos);
        x = (x - scene_pos);
        // multiply that direction by the size (in scene space) and add it to the position
        // this gives us the position of the edge of the point, which we convert in screen space
        vec4 size_vec = $scene_to_framebuffer(scene_pos + normalize(x) * a_size);
        // divide by `w` for perspective, and subtract pos
        // this gives us the actual screen-space size of the point
        $v_size = size_vec.x / size_vec.w - fb_pos.x / fb_pos.w;
        v_edgewidth = ($v_size / a_size) * a_edgewidth;
    }
    else {
        $v_size = a_size * u_px_scale;
        v_edgewidth = a_edgewidth * u_px_scale;
    }

    // gl_PointSize is the diameter
    gl_PointSize = $v_size + 4. * (v_edgewidth + 1.5 * u_antialias);

    if (u_spherical == true) {
        // Get the framebuffer z direction relative to this sphere in visual coords
        vec4 z = $framebuffer_to_visual(fb_pos + vec4(0, 0, big_float, 0));
        z = (z - pos);
        // Get the depth of the sphere in its middle point on the screen
        // size/2 because we need the radius, not the diameter
        vec4 depth_z_vec = $visual_to_framebuffer(pos + normalize(z) * a_size / 2);
        v_depth_middle = depth_z_vec.z / depth_z_vec.w - fb_pos.z / fb_pos.w;
        // size ratio between aliased and non-aliased, needed for correct depth
        v_alias_ratio = gl_PointSize / $v_size;
    }
}
"""


_FRAGMENT_SHADER = """#version 120
uniform vec3 u_light_position;
uniform vec3 u_light_color;
uniform float u_light_ambient;
uniform float u_alpha;
uniform float u_antialias;
uniform bool u_spherical;

varying vec4 v_fg_color;
varying vec4 v_bg_color;
varying float v_edgewidth;
varying float v_depth_middle;
varying float v_alias_ratio;
varying float v_symbol;

void main()
{
    // Discard plotting marker body and edge if zero-size
    if ($v_size <= 0.)
        discard;

    float edgealphafactor = min(v_edgewidth, 1.0);

    float size = $v_size + 4.*(v_edgewidth + 1.5*u_antialias);
    // factor 6 for acute edge angles that need room as for star marker

    // The marker function needs to be linked with this shader
    float r = $marker(gl_PointCoord, size, int(v_symbol));

    // it takes into account an antialising layer
    // of size u_antialias inside the edge
    // r:
    // [-e/2-a, -e/2+a] antialising face-edge
    // [-e/2+a, e/2-a] core edge (center 0, diameter e-2a = 2t)
    // [e/2-a, e/2+a] antialising edge-background
    // use max because we don't want negative transition zone
    float t = max(0.5*v_edgewidth - u_antialias, 0);
    float d = abs(r) - t;

    if (r > 0.5*v_edgewidth + u_antialias)
    {
        // out of the marker (beyond the outer edge of the edge
        // including transition zone due to antialiasing)
        discard;
    }

    vec4 facecolor = v_bg_color;
    vec4 edgecolor = vec4(v_fg_color.rgb, edgealphafactor*v_fg_color.a);
    float depth_change = 0;

    // change color and depth if spherical mode is active
    if (u_spherical == true) {
        // multiply by alias_ratio and then clamp, so we're back to non-alias coordinates
        // and the aliasing ring has the same coordinates as the point just inside,
        // which is important for lighting
        vec2 texcoord = (gl_PointCoord * 2 - 1) * v_alias_ratio;
        float x = clamp(texcoord.x, -1, 1);
        float y = clamp(texcoord.y, -1, 1);
        float z = sqrt(clamp(1 - x*x - y*y, 0, 1));
        vec3 normal = vec3(x, y, z);

        // Diffuse color
        float diffuse = dot(u_light_position, normal);
        // clamp, because 0 < theta < pi/2
        diffuse = clamp(diffuse, 0, 1);
        vec3 diffuse_color = u_light_ambient + u_light_color * diffuse;

        // Specular color
        //   reflect light wrt normal for the reflected ray, then
        //   find the angle made with the eye
        vec3 eye = vec3(0, 0, -1);
        float specular = dot(reflect(u_light_position, normal), eye);
        specular = clamp(specular, 0, 1);
        // raise to the material's shininess, multiply with a
        // small factor for spread
        specular = pow(specular, 80);
        vec3 specular_color = u_light_color * specular;

        facecolor = vec4(facecolor.rgb * diffuse_color + specular_color, facecolor.a * u_alpha);
        edgecolor = vec4(edgecolor.rgb * diffuse_color + specular_color, edgecolor.a * u_alpha);
        // TODO: figure out why this 0.5 is needed, despite already having the radius, not diameter
        depth_change = -0.5 * z * v_depth_middle;
    }

    if (d < 0.0)
    {
        // inside the width of the edge
        // (core, out of the transition zone for antialiasing)
        gl_FragColor = edgecolor;
    }
    else if (v_edgewidth == 0.)
    {// no edge
        if (r > -u_antialias)
        {// outside
            float alpha = 1.0 + r/u_antialias;
            alpha = exp(-alpha*alpha);
            gl_FragColor = vec4(facecolor.rgb, alpha*facecolor.a);
        }
        else
        {// inside
            gl_FragColor = facecolor;
        }
    }
    else
    {// non-zero edge
        float alpha = d/u_antialias;
        alpha = exp(-alpha*alpha);
        if (r > 0.)
        {
            // outer part of the edge: fade out into the background...
            gl_FragColor = vec4(edgecolor.rgb, alpha*edgecolor.a);
        }
        else
        {
            // inner part of the edge: fade into the face color
            gl_FragColor = mix(facecolor, edgecolor, alpha);
        }
    }
    gl_FragDepth = gl_FragCoord.z + depth_change;
}
"""

disc = """
float r = length((pointcoord.xy - vec2(0.5,0.5))*size);
r -= $v_size/2.;
return r;
"""


arrow = """
const float sqrt2 = sqrt(2.);
float half_size = $v_size/2.;
float ady = abs(pointcoord.y -.5)*size;
float dx = (pointcoord.x -.5)*size;
float r1 = abs(dx) + ady - half_size;
float r2 = dx + 0.25*$v_size + ady - half_size;
float r = max(r1,-r2);
return r/sqrt2;//account for slanted edge and correct for width
"""


ring = """
float r1 = length((pointcoord.xy - vec2(0.5,0.5))*size) - $v_size/2.;
float r2 = length((pointcoord.xy - vec2(0.5,0.5))*size) - $v_size/4.;
float r = max(r1,-r2);
return r;
"""

clobber = """
const float sqrt3 = sqrt(3.);
const float PI = 3.14159265358979323846264;
const float t1 = -PI/2;
float circle_radius = 0.32 * $v_size;
float center_shift = 0.36/sqrt3 * $v_size;
//total size (horizontal) = 2*circle_radius + sqrt3*center_shirt = $v_size
vec2  c1 = vec2(cos(t1),sin(t1))*center_shift;
const float t2 = t1+2*PI/3;
vec2  c2 = vec2(cos(t2),sin(t2))*center_shift;
const float t3 = t2+2*PI/3;
vec2  c3 = vec2(cos(t3),sin(t3))*center_shift;
//xy is shift to center marker vertically
vec2 xy = (pointcoord.xy-vec2(0.5,0.5))*size + vec2(0.,-0.25*center_shift);
float r1 = length(xy - c1) - circle_radius;
float r2 = length(xy - c2) - circle_radius;
float r3 = length(xy - c3) - circle_radius;
float r = min(min(r1,r2),r3);
return r;
"""


square = """
float r = max(abs(pointcoord.x -.5)*size, abs(pointcoord.y -.5)*size);
r -= $v_size/2.;
return r;
"""

x = """
vec2 rotcoord = vec2((pointcoord.x + pointcoord.y - 1.) / sqrt(2.),
 (pointcoord.y - pointcoord.x) / sqrt(2.));
//vbar
float r1 = abs(rotcoord.x)*size - $v_size/6.;
float r2 = abs(rotcoord.y)*size - $v_size/2.;
float vbar = max(r1,r2);
//hbar
float r3 = abs(rotcoord.y)*size - $v_size/6.;
float r4 = abs(rotcoord.x)*size - $v_size/2.;
float hbar = max(r3,r4);
return min(vbar, hbar);
"""


diamond = """
float r = abs(pointcoord.x -.5)*size + abs(pointcoord.y -.5)*size;
r -= $v_size/2.;
return r / sqrt(2.);//account for slanted edge and correct for width
"""


vbar = """
float r1 = abs(pointcoord.x - 0.5)*size - $v_size/6.;
float r3 = abs(pointcoord.y - 0.5)*size - $v_size/2.;
float r = max(r1,r3);
return r;
"""

hbar = """
float r2 = abs(pointcoord.y - 0.5)*size - $v_size/6.;
float r3 = abs(pointcoord.x - 0.5)*size - $v_size/2.;
float r = max(r2,r3);
return r;
"""

cross = """
//vbar
float r1 = abs(pointcoord.x - 0.5)*size - $v_size/6.;
float r2 = abs(pointcoord.y - 0.5)*size - $v_size/2.;
float vbar = max(r1,r2);
//hbar
float r3 = abs(pointcoord.y - 0.5)*size - $v_size/6.;
float r4 = abs(pointcoord.x - 0.5)*size - $v_size/2.;
float hbar = max(r3,r4);
return min(vbar, hbar);
"""


tailed_arrow = """
const float sqrt2 = sqrt(2.);
float half_size = $v_size/2.;
float ady = abs(pointcoord.y -.5)*size;
float dx = (pointcoord.x -.5)*size;
float r1 = abs(dx) + ady - half_size;
float r2 = dx + 0.25*$v_size + ady - half_size;
float arrow = max(r1,-r2);
//hbar
float upper_bottom_edges = ady - $v_size/8./sqrt2;
float left_edge = -dx - half_size;
float right_edge = dx + ady - half_size;
float hbar = max(upper_bottom_edges, left_edge);
float scale = 1.; //rescaling for slanted edge
if (right_edge >= hbar)
{
hbar = right_edge;
scale = sqrt2;
}
if (arrow <= hbar)
{
return arrow / sqrt2;//account for slanted edge and correct for width
}
else
{
return hbar / scale;
}
"""


triangle_up = """
float height = $v_size*sqrt(3.)/2.;
float bottom = ((pointcoord.y - 0.5)*size - height/2.);
float rotated_y = sqrt(3.)/2. * (pointcoord.x - 0.5) * size
  - 0.5 * ((pointcoord.y - 0.5)*size - height/6.) + height/6.;
float right_edge = (rotated_y - height/2.);
float cc_rotated_y = -sqrt(3.)/2. * (pointcoord.x - 0.5)*size
  - 0.5 * ((pointcoord.y - 0.5)*size - height/6.) + height/6.;
float left_edge = (cc_rotated_y - height/2.);
float slanted_edges = max(right_edge, left_edge);
return max(slanted_edges, bottom);
"""

triangle_down = """
float height = -$v_size*sqrt(3.)/2.;
float bottom = -((pointcoord.y - 0.5)*size - height/2.);
float rotated_y = sqrt(3.)/2. * (pointcoord.x - 0.5) * size
- 0.5 * ((pointcoord.y - 0.5)*size - height/6.) + height/6.;
float right_edge = -(rotated_y - height/2.);
float cc_rotated_y = -sqrt(3.)/2. * (pointcoord.x - 0.5)*size
- 0.5 * ((pointcoord.y - 0.5)*size - height/6.) + height/6.;
float left_edge = -(cc_rotated_y - height/2.);
float slanted_edges = max(right_edge, left_edge);
return max(slanted_edges, bottom);
"""


star = """
float star = -10000.;
const float PI2_5 = 3.141592653589*2./5.;
const float PI2_20 = 3.141592653589/10.;  //PI*2/20
// downwards shift to that the marker center is halfway vertically
// between the top of the upward spike (y = -v_size/2.)
// and the bottom of one of two downward spikes
// (y = +v_size/2.*cos(2.*pi/10.) approx +v_size/2.*0.8)
// center is at -v_size/2.*0.1
float shift_y = -0.05*$v_size;
// first spike upwards,
// rotate spike by 72 deg four times to complete the star
for (int i = 0; i <= 4; i++)
{
//if not the first spike, rotate it upwards
float x = (pointcoord.x - 0.5)*size;
float y = (pointcoord.y - 0.5)*size;
float spike_rot_angle = float(i) * PI2_5;
float cosangle = cos(spike_rot_angle);
float sinangle = sin(spike_rot_angle);
float spike_x = x;
float spike_y = y + shift_y;
if (i > 0)
{
spike_x = cosangle * x - sinangle * (y + shift_y);
spike_y = sinangle * x + cosangle * (y + shift_y);
}
// in the frame where the spike is upwards:
// rotate 18 deg the zone x < 0 around the top of the star
// (point whose coords are -s/2, 0 where s is the size of the marker)
// compute y coordonates as well because
// we do a second rotation to put the spike at its final position
float rot_center_y = -$v_size/2.;
float rot18x = cos(PI2_20) * spike_x
- sin(PI2_20) * (spike_y - rot_center_y);
//rotate -18 deg the zone x > 0 arount the top of the star
float rot_18x = cos(PI2_20) * spike_x
+ sin(PI2_20) * (spike_y - rot_center_y);
float bottom = spike_y - $v_size/10.;
// max(left edge, right edge)
float spike = max(bottom, max(rot18x, -rot_18x));
if (i == 0)
{// first spike, skip the rotation
star = spike;
}
else // i > 0
{
star = min(star, spike);
}
}
return star;
"""

cross_lines = """
//vbar
float r1 = abs(pointcoord.x - 0.5)*size;
float r2 = abs(pointcoord.y - 0.5)*size - $v_size/2;
float vbar = max(r1,r2);
//hbar
float r3 = abs(pointcoord.y - 0.5)*size;
float r4 = abs(pointcoord.x - 0.5)*size - $v_size/2;
float hbar = max(r3,r4);
return min(vbar, hbar);
"""

symbol_shaders = {
    'disc': disc,
    'arrow': arrow,
    'ring': ring,
    'clobber': clobber,
    'square': square,
    'x': x,
    'diamond': diamond,
    'vbar': vbar,
    'hbar': hbar,
    'cross': cross,
    'tailed_arrow': tailed_arrow,
    'triangle_up': triangle_up,
    'triangle_down': triangle_down,
    'star': star,
    'cross_lines': cross_lines,
}

# combine all the symbol shaders in a big if-else statement
symbol_func = f"""
float symbol(vec2 pointcoord, float size, int symbol) {{
   {' else'.join(
    f''' if (symbol == {i}) {{
        // {name}
        {shader}
    }}'''
    for i, (name, shader) in enumerate(symbol_shaders.items())
    )}
}}"""

# aliases
symbol_aliases = {
    'o': 'disc',
    '+': 'cross',
    '++': 'cross_lines',
    's': 'square',
    '-': 'hbar',
    '|': 'vbar',
    '->': 'tailed_arrow',
    '>': 'arrow',
    '^': 'triangle_up',
    'v': 'triangle_down',
    '*': 'star',
}

symbol_shader_values = {name: i for i, name in enumerate(symbol_shaders)}
symbol_shader_values.update({
    **{alias: symbol_shader_values[name] for alias, name in symbol_aliases.items()},
})


class MarkersVisual(Visual):
    """Visual displaying marker symbols.

    Parameters
    ----------
    pos : array
        The array of locations to display each symbol.
    size : float or array
        The symbol size in screen (or data, if scaling is on) px.
    edge_width : float or array or None
        The width of the symbol outline in screen (or data, if scaling is on) px.
    edge_width_rel : float or array or None
        The width as a fraction of marker size. Exactly one of
        `edge_width` and `edge_width_rel` must be supplied.
    edge_color : Color | ColorArray
        The color used to draw each symbol outline.
    face_color : Color | ColorArray
        The color used to draw each symbol interior.
    symbol : str or array
        The style of symbol used to draw each marker (see Notes).
    scaling : bool
        If set to True, marker scales when rezooming.
    alpha : float
        The opacity level of the visual.
    antialias : float
        Antialiasing amount (in px).
    spherical : bool
        Whether to add a spherical effect on the marker using lighting.
    light_color : Color | ColorArray
        The color of the light used to create the spherical effect.
    light_position : array
        The coordinates of the light used to create the spherical effect.
    light_ambient : float
        The amount of ambient light used to create the spherical effect.

    Notes
    -----
    Allowed style strings are: disc, arrow, ring, clobber, square, diamond,
    vbar, hbar, cross, tailed_arrow, x, triangle_up, triangle_down,
    and star.
    """

    _shaders = {
        'vertex': _VERTEX_SHADER,
        'fragment': _FRAGMENT_SHADER,
    }
    _symbol_shader_values = symbol_shader_values
    _symbol_shader = symbol_func

    def __init__(self, scaling=False, alpha=1, antialias=1, spherical=False,
                 light_color='white', light_position=(1, -1, 1), light_ambient=0.3, **kwargs):
        self._vbo = VertexBuffer()
        self._data = None

        Visual.__init__(self, vcode=self._shaders['vertex'], fcode=self._shaders['fragment'])
        self._symbol_func = Function(self._symbol_shader)
        self.shared_program.frag['marker'] = self._symbol_func
        self._v_size_var = Variable('varying float v_size')
        self.shared_program.vert['v_size'] = self._v_size_var
        self.shared_program.frag['v_size'] = self._v_size_var
        self._symbol_func['v_size'] = self._v_size_var

        self.set_gl_state(depth_test=True, blend=True,
                          blend_func=('src_alpha', 'one_minus_src_alpha'))
        self._draw_mode = 'points'

        if len(kwargs) > 0:
            self.set_data(**kwargs)

        self.scaling = scaling
        self.antialias = antialias
        self.light_color = light_color
        self.light_position = light_position
        self.light_ambient = light_ambient
        self.alpha = alpha
        self.spherical = spherical

        self.freeze()

    def set_data(self, pos=None, size=10., edge_width=1., edge_width_rel=None,
                 edge_color='black', face_color='white',
                 symbol='o'):
        """Set the data used to display this visual.

        Parameters
        ----------
        pos : array
            The array of locations to display each symbol.
        size : float or array
            The symbol size in screen (or data, if scaling is on) px.
        edge_width : float or array or None
            The width of the symbol outline in screen (or data, if scaling is on) px.
        edge_width_rel : float or array or None
            The width as a fraction of marker size. Exactly one of
            `edge_width` and `edge_width_rel` must be supplied.
        edge_color : Color | ColorArray
            The color used to draw each symbol outline.
        face_color : Color | ColorArray
            The color used to draw each symbol interior.
        symbol : str or array
            The style of symbol used to draw each marker (see Notes).
        """
        if (edge_width is not None) + (edge_width_rel is not None) != 1:
            raise ValueError('exactly one of edge_width and edge_width_rel '
                             'must be non-None')

        if edge_width is not None:
            edge_width = np.asarray(edge_width)
            if np.any(edge_width < 0):
                raise ValueError('edge_width cannot be negative')
        else:
            edge_width_rel = np.asarray(edge_width_rel)
            if np.any(edge_width_rel < 0):
                raise ValueError('edge_width_rel cannot be negative')

        if symbol is not None:
            if not np.all(np.isin(np.asarray(symbol), self.symbols)):
                raise ValueError(f'symbols must one of {self.symbols}')

        edge_color = ColorArray(edge_color).rgba
        if len(edge_color) == 1:
            edge_color = edge_color[0]

        face_color = ColorArray(face_color).rgba
        if len(face_color) == 1:
            face_color = face_color[0]

        if pos is not None:
            assert (isinstance(pos, np.ndarray) and
                    pos.ndim == 2 and pos.shape[1] in (2, 3))

            n = len(pos)
            data = np.zeros(n, dtype=[('a_position', np.float32, 3),
                                      ('a_fg_color', np.float32, 4),
                                      ('a_bg_color', np.float32, 4),
                                      ('a_size', np.float32),
                                      ('a_edgewidth', np.float32),
                                      ('a_symbol', np.float32)])
            data['a_fg_color'] = edge_color
            data['a_bg_color'] = face_color
            if edge_width is not None:
                data['a_edgewidth'] = edge_width
            else:
                data['a_edgewidth'] = size * edge_width_rel
            data['a_position'][:, :pos.shape[1]] = pos
            data['a_size'] = size

            data['a_symbol'] = np.vectorize(self._symbol_shader_values.get)(symbol)

            self._data = data
            self._vbo.set_data(data)
            self.shared_program.bind(self._vbo)

        self.update()

    @property
    def symbols(self):
        return list(self._symbol_shader_values)

    @property
    def symbol(self):
        if self._data is None:
            return None
        value_to_symbol = {v: k for k, v in self._symbol_shader_values.items()}
        return np.vectorize(value_to_symbol.get)(self._data['a_symbol'])

    @symbol.setter
    def symbol(self, value):
        if self._data is not None:
            rec_to_kw = {
                'a_position': 'pos',
                'a_fg_color': 'edge_color',
                'a_bg_color': 'face_color',
                'a_size': 'size',
                'a_edgewidth': 'edge_width',
                'a_symbol': 'symbol',
            }
            kwargs = {kw: self._data[rec] for rec, kw in rec_to_kw.items()}
        else:
            kwargs = {}
        kwargs['symbol'] = value
        self.set_data(**kwargs)

    @property
    def scaling(self):
        """
        If set to True, marker scales when rezooming.
        """
        return self._scaling

    @scaling.setter
    def scaling(self, value):
        value = bool(value)
        self.shared_program['u_scaling'] = value
        self._scaling = value
        self.update()

    @property
    def antialias(self):
        """
        Antialiasing amount (in px).
        """
        return self._antialias

    @antialias.setter
    def antialias(self, value):
        value = float(value)
        self.shared_program['u_antialias'] = value
        self._antialias = value
        self.update()

    @property
    def light_position(self):
        """
        The coordinates of the light used to create the spherical effect.
        """
        return self._light_position

    @light_position.setter
    def light_position(self, value):
        value = np.array(value)
        self.shared_program['u_light_position'] = value / np.linalg.norm(value)
        self._light_position = value
        self.update()

    @property
    def light_ambient(self):
        """
        The amount of ambient light used to create the spherical effect.
        """
        return self._light_ambient

    @light_ambient.setter
    def light_ambient(self, value):
        self.shared_program['u_light_ambient'] = value
        self._light_ambient = value
        self.update()

    @property
    def light_color(self):
        """
        The color of the light used to create the spherical effect.
        """
        return self._light_color

    @light_color.setter
    def light_color(self, value):
        self.shared_program['u_light_color'] = ColorArray(value).rgb
        self._light_color = value
        self.update()

    @property
    def alpha(self):
        """
        The opacity level of the visual.
        """
        return self._alpha

    @alpha.setter
    def alpha(self, value):
        self.shared_program['u_alpha'] = value
        self._alpha = value
        self.update()

    @property
    def spherical(self):
        """
        Whether to add a spherical effect on the marker using lighting.
        """
        return self._spherical

    @spherical.setter
    def spherical(self, value):
        self.shared_program['u_spherical'] = value
        self._spherical = value
        self.update()

    def _prepare_transforms(self, view):
        view.view_program.vert['visual_to_framebuffer'] = view.get_transform('visual', 'framebuffer')
        view.view_program.vert['framebuffer_to_visual'] = view.get_transform('framebuffer', 'visual')
        view.view_program.vert['framebuffer_to_render'] = view.get_transform('framebuffer', 'render')
        view.view_program.vert['framebuffer_to_scene'] = view.get_transform('framebuffer', 'scene')
        view.view_program.vert['scene_to_framebuffer'] = view.get_transform('scene', 'framebuffer')

    def _prepare_draw(self, view):
        if self._data is None:
            return False
        view.view_program['u_px_scale'] = view.transforms.pixel_scale
        view.view_program['u_scaling'] = self.scaling

    def _compute_bounds(self, axis, view):
        pos = self._data['a_position']
        if pos is None:
            return None
        if pos.shape[1] > axis:
            return (pos[:, axis].min(), pos[:, axis].max())
        else:
            return (0, 0)



# -*- coding: utf-8 -*-

import pytest
import numpy as np
from vispy import scene

from vispy.testing import (TestingCanvas, requires_application,
                           run_tests_if_main, requires_pyopengl,
                           raises)
from vispy.testing.image_tester import assert_image_approved, downsample
from vispy.testing.rendered_array_tester import compare_render, max_for_dtype


@requires_pyopengl()
def test_volume():
    vol = np.zeros((20, 20, 20), 'float32')
    vol[8:16, 8:16, :] = 1.0

    # Create
    V = scene.visuals.Volume(vol)
    assert V.clim == (0, 1)
    assert V.method == 'mip'
    assert V.interpolation == 'linear'

    # Set wrong data
    with raises(ValueError):
        V.set_data(np.zeros((20, 20), 'float32'))

    # Clim
    V.set_data(vol, (0.5, 0.8))
    assert V.clim == (0.5, 0.8)
    with raises(ValueError):
        V.set_data((0.5, 0.8, 1.0))

    # Method
    V.method = 'iso'
    assert V.method == 'iso'

    # Interpolation
    V.interpolation = 'nearest'
    assert V.interpolation == 'nearest'

    # Step size
    V.relative_step_size = 1.1
    assert V.relative_step_size == 1.1
    # Disallow 0 step size to avoid GPU stalling
    with raises(ValueError):
        V.relative_step_size = 0


@requires_pyopengl()
def test_volume_bounds():
    vol = np.zeros((20, 30, 40), 'float32')
    vol[8:16, 8:16, :] = 1.0

    # Create
    V = scene.visuals.Volume(vol)
    assert V._compute_bounds(0, V) == (0, 40)  # x
    assert V._compute_bounds(1, V) == (0, 30)  # y
    assert V._compute_bounds(2, V) == (0, 20)  # z


@requires_pyopengl()
@requires_application()
def test_volume_draw():
    with TestingCanvas(bgcolor='k', size=(100, 100)) as c:
        v = c.central_widget.add_view()
        v.camera = 'turntable'
        v.camera.fov = 70
        # Create
        np.random.seed(2376)
        vol = np.random.normal(size=(20, 20, 20), loc=0.5, scale=0.2)
        vol[8:16, 8:16, :] += 1.0
        scene.visuals.Volume(vol, parent=v.scene)  # noqa
        v.camera.set_range()
        assert_image_approved(c.render(), 'visuals/volume.png')


@requires_pyopengl()
@requires_application()
@pytest.mark.parametrize('clim_on_init', [False, True])
@pytest.mark.parametrize('texture_format', [None, '__dtype__', 'auto'])
@pytest.mark.parametrize('input_dtype', [np.uint8, np.uint16, np.float32, np.float64])
def test_volume_clims_and_gamma(texture_format, input_dtype, clim_on_init):
    """Test volume visual with clims and gamma on shader.

    Test is parameterized based on ``texture_format`` and should produce
    relatively the same results for each format.

    Currently just using np.ones since the angle of view made more complicated samples
    challenging, but this confirms gamma and clims works in the shader.
    The VolumeVisual defaults to the "grays" colormap so although we compare
    data using RGBA arrays, each R/G/B channel should be the same.

    """
    size = (40, 40)
    if texture_format == '__dtype__':
        texture_format = input_dtype
    np.random.seed(0)  # make tests the same every time
    data = _make_test_data(size[:1] * 3, input_dtype)
    clim = (0, 1)
    new_clim = (0.3, 0.8)
    max = max_for_dtype(input_dtype)
    if max != 1:
        clim = (clim[0] * max, clim[1] * max)
        new_clim = (new_clim[0] * max, new_clim[1] * max)

    kwargs = {}
    if clim_on_init:
        kwargs['clim'] = clim
    with TestingCanvas(size=size, bgcolor="k") as c:
        v = c.central_widget.add_view(border_width=0)
        volume = scene.visuals.Volume(
            data,
            interpolation='nearest',
            parent=v.scene,
            method='mip',
            texture_format=texture_format,
            **kwargs
        )
        v.camera = 'arcball'
        v.camera.fov = 0
        v.camera.scale_factor = 40.0
        v.camera.center = (19.5, 19.5, 19.5)

        rendered = c.render()
        _dtype = rendered.dtype
        shape_ratio = rendered.shape[0] // data.shape[0]
        rendered1 = downsample(rendered, shape_ratio, axis=(0, 1)).astype(_dtype)
        predicted = data.max(axis=1)
        compare_render(predicted, rendered1)

        # adjust contrast limits
        volume.clim = new_clim
        rendered2 = downsample(c.render(), shape_ratio, axis=(0, 1)).astype(_dtype)
        scaled_data = (np.clip(data, new_clim[0], new_clim[1]) - new_clim[0]) / (new_clim[1] - new_clim[0])
        predicted = scaled_data.max(axis=1)
        compare_render(predicted, rendered2, previous_render=rendered)

        # adjust gamma
        volume.gamma = 2
        rendered3 = downsample(c.render(), shape_ratio, axis=(0, 1)).astype(_dtype)
        predicted = (scaled_data ** 2).max(axis=1)
        compare_render(predicted, rendered3, previous_render=rendered2)


@requires_pyopengl()
@requires_application()
@pytest.mark.parametrize('method_name', scene.visuals.Volume._rendering_methods.keys())
def test_all_render_methods(method_name):
    """Test that render methods don't produce any errors."""
    size = (40, 40)
    np.random.seed(0)  # make tests the same every time
    data = _make_test_data(size[:1] * 3, np.float32)
    # modify the data for 'minip' method so that there is at least one segment
    # of the volume with no 'empty'/zero space
    data[:, :, 40 // 3: 2 * 40 // 3] = 1.0
    clim = (0, 1)
    kwargs = {}
    with TestingCanvas(size=size, bgcolor="k") as c:
        v = c.central_widget.add_view(border_width=0)
        volume = scene.visuals.Volume(
            data,
            interpolation='nearest',
            clim=clim,
            parent=v.scene,
            method=method_name,
            **kwargs
        )
        v.camera = 'arcball'
        v.camera.fov = 0
        v.camera.scale_factor = 40.0
        v.camera.center = (19.5, 19.5, 19.5)

        assert volume.method == method_name
        rendered = c.render()[..., :3]
        # not all black
        assert rendered.sum() != 0
        # not all white
        assert rendered.sum() != 255 * rendered.size


@requires_pyopengl()
@requires_application()
@pytest.mark.parametrize('texture_format', [None, 'auto'])
def test_equal_clims(texture_format):
    """Test that equal clims produce a min cmap value."""
    size = (40, 40)
    np.random.seed(0)  # make tests the same every time
    data = _make_test_data(size[:1] * 3, np.float32)
    with TestingCanvas(size=size, bgcolor="k") as c:
        v = c.central_widget.add_view(border_width=0)
        scene.visuals.Volume(
            data,
            interpolation='nearest',
            clim=(128.0, 128.0),  # equal clims
            cmap='viridis',  # something with a non-black min value
            parent=v.scene,
            method='mip',
            texture_format=texture_format,
        )
        v.camera = 'arcball'
        v.camera.fov = 0
        v.camera.scale_factor = 40.0
        v.camera.center = (19.5, 19.5, 19.5)

        rendered = c.render()[..., :3]
        # not all black
        assert rendered.sum() != 0
        # not all white
        assert rendered.sum() != 255 * rendered.size
        # should be all the same value
        r_unique = np.unique(rendered[..., 0])
        g_unique = np.unique(rendered[..., 1])
        b_unique = np.unique(rendered[..., 2])
        assert r_unique.size == 1
        assert g_unique.size == 1
        assert b_unique.size == 1


def _make_test_data(shape, input_dtype):
    one_third = shape[0] // 3
    two_third = 2 * one_third
    data = np.zeros(shape, dtype=np.float64)
    # 0.00 | 1.00 | 0.00
    # 0.50 | 0.00 | 0.25
    # 0.00 | 0.00 | 0.00
    data[:, :one_third, one_third:two_third] = 1.0
    data[:, one_third:two_third, :one_third] = 0.5
    data[:, one_third:two_third, two_third:] = 0.25
    max_val = max_for_dtype(input_dtype)
    if max_val != 1:
        data *= max_val
    data = data.astype(input_dtype)
    return data


@requires_pyopengl()
def test_set_data_does_not_change_input():
    # Create volume
    V = scene.visuals.Volume(np.zeros((20, 20, 20), dtype=np.float32))

    # calling Volume.set_data() should NOT alter the values of the input array
    # regardless of data type
    vol = np.random.randint(0, 200, (20, 20, 20))
    for dtype in ['uint8', 'int16', 'uint16', 'float32', 'float64']:
        vol_copy = np.array(vol, dtype=dtype, copy=True)
        # setting clim so that normalization would otherwise change the data
        V.set_data(vol_copy, clim=(0, 200))
        assert np.allclose(vol, vol_copy)

    # dtype has to be the same as the one used to init the texture, or it will
    # be first coerced to the same dtype as the init

    vol2 = np.array(vol, dtype=np.float32, copy=True)
    assert np.allclose(vol, vol2)
    # we explicitly create a copy when data would be altered by the texture,
    # no matter what the user asks, so the data outside should never change
    V.set_data(vol2, clim=(0, 200), copy=False)
    assert np.allclose(vol, vol2)


@requires_pyopengl()
def test_set_data_changes_shape():
    dtype = np.float32
    # Create initial volume
    V = scene.visuals.Volume(np.zeros((20, 20, 20), dtype=dtype))

    # Sending new three dimensional data of different shape should alter volume shape
    vol = np.zeros((25, 25, 10), dtype=dtype)
    V.set_data(vol)
    assert V._vol_shape == (25, 25, 10)

    # Sending data of dimension other than 3 should raise a ValueError
    vol2 = np.zeros((20, 20), dtype=dtype)
    with pytest.raises(ValueError):
        V.set_data(vol2)

    vol2 = np.zeros((20, 20, 20, 20), dtype=dtype)
    with pytest.raises(ValueError):
        V.set_data(vol2)


@requires_pyopengl()
@requires_application()
def test_changing_cmap():
    """Test that changing colormaps updates the display."""
    size = (40, 40)
    np.random.seed(0)  # make tests the same every time
    data = _make_test_data(size[:1] * 3, np.float32)
    cmap = 'grays'
    test_cmaps = ('reds', 'greens', 'blues')
    clim = (0, 1)
    kwargs = {}
    with TestingCanvas(size=size, bgcolor="k") as c:
        v = c.central_widget.add_view(border_width=0)
        volume = scene.visuals.Volume(
            data,
            interpolation='nearest',
            clim=clim,
            cmap=cmap,
            parent=v.scene,
            **kwargs
        )
        v.camera = 'arcball'
        v.camera.fov = 0
        v.camera.scale_factor = 40.0

        # render with grays colormap
        grays = c.render()

        # update cmap, compare rendered array with the grays cmap render
        for cmap in test_cmaps:
            volume.cmap = cmap
            current_cmap = c.render()
            with pytest.raises(AssertionError):
                np.testing.assert_allclose(grays, current_cmap)


@requires_pyopengl()
@requires_application()
def test_plane_depth():
    with TestingCanvas(size=(80, 80)) as c:
        v = c.central_widget.add_view(border_width=0)
        v.camera = 'arcball'
        v.camera.fov = 0
        v.camera.center = (40, 40, 40)
        v.camera.scale_factor = 80.0

        # two planes at 45 degrees relative to the camera. If depth is set correctly, we should see one half
        # of the screen red and the other half white
        scene.visuals.Volume(
            np.ones((80, 80, 80), dtype=np.uint8),
            interpolation="nearest",
            clim=(0, 1),
            cmap="grays",
            raycasting_mode="plane",
            plane_normal=(0, 1, 1),
            parent=v.scene,
        )

        scene.visuals.Volume(
            np.ones((80, 80, 80), dtype=np.uint8),
            interpolation="nearest",
            clim=(0, 1),
            cmap="reds",
            raycasting_mode="plane",
            plane_normal=(0, 1, -1),
            parent=v.scene,
        )

        # render with grays colormap
        rendered = c.render()
        left = rendered[40, 20]
        right = rendered[40, 60]
        assert np.array_equal(left, [255, 0, 0, 255])
        assert np.array_equal(right, [255, 255, 255, 255])


@requires_pyopengl()
@requires_application()
def test_volume_depth():
    """Check that depth setting is properly performed for the volume visual

    Render a volume with a blue ball in front of a red plane in front of a
    blue plane, checking that the output image contains both red and blue pixels.
    """
    # A blue strip behind a red strip
    # If depth is set correctly, we should see only red pixels
    # the screen
    blue_vol = np.zeros((80, 80, 80), dtype=np.uint8)
    blue_vol[:, -1, :] = 1  # back plane blue
    blue_vol[30:50, 30:50, 30:50] = 1  # blue in center

    red_vol = np.zeros((80, 80, 80), dtype=np.uint8)
    red_vol[:, -5, :] = 1  # red plane in front of blue plane

    with TestingCanvas(size=(80, 80)) as c:
        v = c.central_widget.add_view(border_width=0)
        v.camera = 'arcball'
        v.camera.fov = 0
        v.camera.center = (40, 40, 40)
        v.camera.scale_factor = 80.0

        scene.visuals.Volume(
            red_vol,
            interpolation="nearest",
            clim=(0, 1),
            cmap="reds",
            parent=v.scene,
        )

        scene.visuals.Volume(
            blue_vol,
            interpolation="nearest",
            clim=(0, 1),
            cmap="blues",
            parent=v.scene,
        )

        # render
        rendered = c.render()
        reds = np.sum(rendered[:, :, 0])
        greens = np.sum(rendered[:, :, 1])
        blues = np.sum(rendered[:, :, 2])
        assert reds > 0
        np.testing.assert_allclose(greens, 0)
        assert blues > 0


@requires_pyopengl()
@requires_application()
def test_mip_cutoff():
    """
    Ensure fragments are properly discarded based on the mip_cutoff
    for the mip and attenuated_mip rendering methods
    """
    with TestingCanvas(size=(80, 80)) as c:
        v = c.central_widget.add_view(border_width=0)
        v.camera = 'arcball'
        v.camera.fov = 0
        v.camera.center = (40, 40, 40)
        v.camera.scale_factor = 80.0

        vol = scene.visuals.Volume(
            np.ones((80, 80, 80), dtype=np.uint8),
            interpolation="nearest",
            clim=(0, 1),
            cmap="grays",
            parent=v.scene,
        )

        # we should see white
        rendered = c.render()
        assert np.array_equal(rendered[40, 40], [255, 255, 255, 255])

        vol.mip_cutoff = 10
        # we should see black
        rendered = c.render()
        assert np.array_equal(rendered[40, 40], [0, 0, 0, 255])

        # repeat for attenuated_mip
        vol.method = 'attenuated_mip'
        vol.mip_cutoff = None

        # we should see white
        rendered = c.render()
        assert np.array_equal(rendered[40, 40], [255, 255, 255, 255])

        vol.mip_cutoff = 10
        # we should see black
        rendered = c.render()
        assert np.array_equal(rendered[40, 40], [0, 0, 0, 255])


@requires_pyopengl()
@requires_application()
def test_minip_cutoff():
    """
    Ensure fragments are properly discarded based on the minip_cutoff
    for the minip rendering method
    """
    with TestingCanvas(size=(80, 80)) as c:
        v = c.central_widget.add_view(border_width=0)
        v.camera = 'arcball'
        v.camera.fov = 0
        v.camera.center = (40, 40, 40)
        v.camera.scale_factor = 120.0

        # just surface of the cube is ones, but it should win over the twos inside
        data = np.ones((80, 80, 80), dtype=np.uint8)
        data[1:-1, 1:-1, 1:-1] = 2

        vol = scene.visuals.Volume(
            data,
            interpolation="nearest",
            method='minip',
            clim=(0, 2),
            cmap="grays",
            parent=v.scene,
        )

        # we should see gray (half of cmap)
        rendered = c.render()
        assert np.array_equal(rendered[40, 40], [128, 128, 128, 255])

        # discard fragments above -10 (everything)
        vol.minip_cutoff = -10
        # we should see black
        rendered = c.render()
        assert np.array_equal(rendered[40, 40], [0, 0, 0, 255])


@requires_pyopengl()
@requires_application()
def test_volume_set_data_different_dtype():
    size = (80, 80)
    data = np.array([[[0, 127]]], dtype=np.int8)
    left = (40, 10)
    right = (40, 70)
    white = (255, 255, 255, 255)
    black = (0, 0, 0, 255)

    with TestingCanvas(size=size[::-1], bgcolor="w") as c:
        view = c.central_widget.add_view()
        view.camera = 'arcball'
        view.camera.fov = 0
        view.camera.center = 0.5, 0, 0
        view.camera.scale_factor = 2
        volume = scene.visuals.Volume(
            data,
            cmap='grays',
            clim=[0, 127],
            parent=view.scene
        )

        render = c.render()
        assert np.allclose(render[left], black)
        assert np.allclose(render[right], white)

        # same data as float should change nothing
        volume.set_data(data.astype(np.float32))
        render = c.render()
        assert np.allclose(render[left], black)
        assert np.allclose(render[right], white)

        # something inverted, different dtype
        new_data = np.array([[[127, 0]]], dtype=np.float16)
        volume.set_data(new_data)
        render = c.render()
        assert np.allclose(render[left], white)
        assert np.allclose(render[right], black)

        # out of bounds should clip (2000 > 127)
        new_data = np.array([[[0, 2000]]], dtype=np.float64)
        volume.set_data(new_data)
        render = c.render()
        assert np.allclose(render[left], black)
        assert np.allclose(render[right], white)


run_tests_if_main()



# -*- coding: utf-8 -*-
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

"""
ptime.py -  Precision time function made os-independent
"""

import time as systime
# get a reference starting time - initial performance counter
START_TIME = systime.time() - systime.perf_counter()


def time():
    # return reference starting time + delta of performance counters
    return START_TIME + systime.perf_counter()
